import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats


def setup_canvas(nx, ny, figsize = (10, 6), sharex = True, sharey = True, markersize = 2, fontsize = 16, flatten = True, labelsize= 15, wspace = 0, hspace = 0, panels = False):
    plt.rcParams.update({'lines.markersize': markersize, 'font.size': fontsize})
    fig, axes = plt.subplots(nx, ny, figsize = figsize, sharex = sharex, sharey = sharey)
    if nx * ny == 1: axes = np.array([axes])
    if flatten: axes = axes.flatten()
    for ax in axes.flatten():
        ax.tick_params(direction = "in", which = "both", labelsize = labelsize)

    if panels:
        for i in range(len(axes)):
            axes[i].text(0.05, 0.8, f'({chr(97 + i)})', transform = axes[i].transAxes)

    plt.subplots_adjust(wspace = wspace, hspace = hspace)

    if len(axes) == 1:
        return fig, axes[0]
    else:
        return fig, axes

def upper_legend(ax, xloc = 0.5, yloc = 1.1, ncols = None, nrows = None, user_labels = [], user_labels_order = [], loc = "upper center", framealpha = 0., frameon = False):
    def reorder(list_in, nrows):
        ncols = len(list_in) // nrows
        if nrows * ncols != len(list_in): ncols += 1
        list_out = []
        for c in range(ncols):
            for r in range(nrows):
                if r * ncols + c >= len(list_in): continue
                list_out.append(list_in[r * ncols + c])
        assert len(list_in) == len(list_out), 'ERROR: len in != len out'
        return list_out, ncols
    handles, labels = ax.get_legend_handles_labels()
    if user_labels: labels = user_labels
    if user_labels_order:
        handles = [dict(zip(labels, handles))[l] for l in user_labels_order]
        labels = user_labels_order
    if len(handles) != len(labels): print('WARNING: the lengths are unequal')
    if nrows:
        labels, ncols = reorder(labels, nrows)
        handles, ncols = reorder(handles, nrows)
    by_label = dict(zip(labels, handles))
    if not ncols: ncols = len(labels)

    x0 = ax.get_position().x0
    x1 = ax.get_position().x1
    y0 = ax.get_position().y0
    y1 = ax.get_position().y1

    if xloc and yloc:
        ax.legend(by_label.values(), by_label.keys(), loc = loc, framealpha = framealpha, frameon = frameon, bbox_to_anchor=(xloc, yloc), ncol=ncols)
    else:
        ax.legend(by_label.values(), by_label.keys(), loc = loc, framealpha = framealpha, frameon = frameon, ncol=ncols, bbox_to_anchor=(0., -0.05, 1., 0.), borderaxespad=0, mode='expand')
    return ax

def nticks_prune(ax, which = 'x', nbins = None, prune = None):
    # prune: can be upper
    if which == 'x':
        if not nbins:
            nbins = len(ax.get_xticklabels()) # added 
        ax.xaxis.set_major_locator(plt.MaxNLocator(nbins = nbins, prune = prune))
    else:
        if not nbins:
            nbins = len(ax.get_yticklabels()) # added 
        ax.yaxis.set_major_locator(plt.MaxNLocator(nbins = nbins, prune = prune)) 
    return ax

def rotate_ticks(ax, which, degree):
    ax.tick_params(axis=which, rotation=degree)

def sort_list_by(lista, listb):
    # sort lista by listb
    lista = [i for _, i in sorted(zip(listb, lista))]
    return lista

def add_text(ax, x, y, text, horizontalalignment = 'center', verticalalignment = 'center'):
    ax.text(x, y, text, transform = ax.transAxes, horizontalalignment = horizontalalignment, verticalalignment = verticalalignment)

def add_line(ax, loc, linestyle = '--', color = 'k', alpha = 0.5, direction = 'h', bmin = 0, bmax = 1):
    if direction.lower() in ['h', 'horizontal']:
        ax.axhline(loc, linestyle = linestyle, color = color, alpha = alpha, xmin = bmin, xmax = bmax)
    elif direction.lower() in ['v', 'vertical']:
        ax.axvline(loc, linestyle = linestyle, color = color, alpha = alpha, ymin = bmin, ymax = bmax)
    else:
        raise Exception('Options: h, horizontal, v, and vertical!')

def shift_axis_label(ax, which, x_shift, y_shift):
    if which == 'x':
        ax.xaxis.set_label_coords(x_shift, y_shift, transform = ax.transAxes)
    elif which == 'y':
        ax.yaxis.set_label_coords(x_shift, y_shift, transform = ax.transAxes)
    else:
        raise Exception('wrong axis')

def init_sci_env(fontsize = 14, linemarkersize = 2, legendtitle_fontsize = 14, figuresize = (10, 6), pandas_max_columns = None):
    # plt.rcParams["figure.figsize"] = (10, 6)
    pd.set_option('display.max_columns', pandas_max_columns)
    plt.rcParams.update({
        "font.size": fontsize, 
        "lines.markersize": linemarkersize, 
        'legend.title_fontsize': legendtitle_fontsize, 
        "figure.figsize": figuresize
    })
    colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
    return colors

def reset_sci_env():
    pd.reset_option('all')
    plt.rcParams.update(plt.rcParamsDefault)

def savefig(fig, savefile, dpi = 600, bbox_inches = "tight", transparent = False, **kwargs):
    fig.savefig(savefile, dpi = dpi, bbox_inches = bbox_inches, transparent = transparent, **kwargs)


from scipy.stats import gaussian_kde

def kde_scatter(ax, dfp, x_name, y_name, frac = 0.3, v_scale = 0.1, vmin = None, vmax = None, cmap = 'RdYlBu_r'):
    dfp = dfp[[x_name, y_name]].dropna().sample(frac = frac).reset_index(drop = True)
    x = dfp[x_name]
    y = dfp[y_name]

    # Calculate the point density
    xy = np.vstack([x,y])
    z = gaussian_kde(xy)(xy)

    # Sort the points by density, so that the densest points are plotted last
    idx = z.argsort()
    x, y, z = x[idx], y[idx], z[idx]

    # fig, ax = plt.subplots(1, 1, figsize = (9, 9))
    ax.scatter(x, y, c=z, s=50, vmin = vmin, vmax = vmax, cmap = cmap)

    xl = np.linspace(np.floor(x.min()), np.ceil(x.max()), 1000)
    ax.plot(xl, xl, ls = '-.', color = 'k')

    x_min = ax.get_xlim()[0]
    x_max = ax.get_xlim()[1]
    y_min = ax.get_ylim()[0]
    y_max = ax.get_ylim()[1]
    v_min = np.min([x_min, y_min])
    v_max = np.max([x_max, y_max])
    v_ran = v_max - v_min

    ax.set_xlim(v_min - v_ran * v_scale, v_max +  v_ran * v_scale)
    ax.set_ylim(v_min - v_ran * v_scale, v_max +  v_ran * v_scale)

def line_dot(df, ax, colors = None, legend = True, markercolor = 'white', edgecolor = 'k', markersize = 20, zorder = 10):
    if not colors: colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
    df.plot(ax = ax, style = '--', color = colors, legend = legend)
    for c in df.columns:
        ax.scatter(df.index, df[c], color = markercolor, edgecolor = edgecolor, s = markersize, zorder = zorder)
    return ax

def custom_cmap(clist, cname = 'custom_cmap', N = 256):
    import matplotlib.colors
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list(cname, clist, N=N)
    return cmap

colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
nature_colors = ['#E64B35B2', '#4DBBD5B2', '#00A087B2', '#3C5488B2', '#F39B7FB2', '#8491B4B2', '#91D1C2B2', '#DC0000B2', '#7E6148B2', '#B09C85B2'] # alpha = 0.7
nature_colors_01 = ['#E64B3519', '#4DBBD519', '#00A08719', '#3C548819', '#F39B7F19', '#8491B419', '#91D1C219', '#DC000019', '#7E614819', '#B09C8519'] # alpha = 0.1
nature_colors_03 = ['#E64B354C', '#4DBBD54C', '#00A0874C', '#3C54884C', '#F39B7F4C', '#8491B44C', '#91D1C24C', '#DC00004C', '#7E61484C', '#B09C854C'] # alpha = 0.3
nature_colors_05 = ['#E64B357F', '#4DBBD57F', '#00A0877F', '#3C54887F', '#F39B7F7F', '#8491B47F', '#91D1C27F', '#DC00007F', '#7E61487F', '#B09C857F'] # alpha = 0.5
nature_colors_07 = ['#E64B35B2', '#4DBBD5B2', '#00A087B2', '#3C5488B2', '#F39B7FB2', '#8491B4B2', '#91D1C2B2', '#DC0000B2', '#7E6148B2', '#B09C85B2'] # alpha = 0.7
nature_colors_09 = ['#E64B35E5', '#4DBBD5E5', '#00A087E5', '#3C5488E5', '#F39B7FE5', '#8491B4E5', '#91D1C2E5', '#DC0000E5', '#7E6148E5', '#B09C85E5'] # alpha = 0.9
nature_colors_10 = ['#E64B35FF', '#4DBBD5FF', '#00A087FF', '#3C5488FF', '#F39B7FFF', '#8491B4FF', '#91D1C2FF', '#DC0000FF', '#7E6148FF', '#B09C85FF'] # alpha = 1.0

nature_colors_base = {
    'Cinnabar': '#E64B35', 
    'Shakespeare': '#4DBBD5', 
    'PersianGreen': '#00A087',
    'Chambray': '#3C5488',
    'Apricot': '#F39B7F',
    'WildBlueYonder': '#8491B4',
    'MonteCarlo': '#91D1C2',
    'Monza': '#DC0000',
    'RomanCoffee': '#7E6148',
    'Sandrift': '#B09C85',
}


def show_colors(colors, width = 1, height = 1, hspace = 0.05, wspace = 0.05, fontsize = 10):
    import matplotlib.patches
    import matplotlib.collections

    ncolors = len(colors)
    nrow = ncol = np.floor(np.sqrt(ncolors))
    nrow = int(nrow); ncol = int(ncol)
    while nrow * ncol < ncolors:
        nrow = nrow + 1

    xx = np.arange(0, ncol, (width + wspace))
    yy = np.arange(0, nrow, (height + hspace))

    fig, ax = setup_canvas(1, 1, figsize = (6, 6))

    patches = []
    for i, xi in enumerate(xx):
        for j, yi in enumerate(yy):
            cnt = i * nrow + j
            if cnt < ncolors:
                text = colors[cnt]
                color = colors[cnt]
            else:
                text = None
                color = 'None'
            patch = matplotlib.patches.Rectangle((xi, yi), width, height, fill = True, color = color)
            ax.add_patch(patch)
            ax.text(xi + width / 2, yi  + height / 2, text, fontsize = fontsize, horizontalalignment = 'center', verticalalignment = 'center')

    pc = matplotlib.collections.PatchCollection(patches)
    ax.add_collection(pc)

    ax.relim()
    ax.autoscale_view()

    ax.axis("off")
    return fig, ax


'''
Basic scientific computing import
'''
import os, sys, subprocess, shutil, warnings, pickle, time, gc
from copy import deepcopy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path
from scipy import stats
from tqdm import tqdm
from datetime import datetime, timedelta

def install(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])

try:
    import rasterio as rio
except Exception as e:
    print(e)
    install('rasterio')
    import rasterio as rio
try:
    import geopandas as gpd
except Exception as e:
    print(e)
    install('geopandas')
    import geopandas as gpd
try:
    import xarray as xr
except Exception as e:
    print(e)
    install('xarray')
    install('netCDF4')
    import xarray as xr
try:
    import rioxarray as rxr
except Exception as e:
    print(e)
    install('rioxarray')
    import rioxarray as rxr
try:
    import seaborn as sns
except Exception as e:
    print(e)
    install('seaborn')
    import seaborn as sns
import pip
from operator import imod

# # ------------------------------------------------------------------------------
# def option_load(opt = 0):
#     def install(package):
#         subprocess.check_call([sys.executable, "-m", "pip", "install", package])
#     def geo_load():
#         global rio, gpd, xr, rxr
#         try:
#             import rasterio as rio
#         except Exception as e:
#             print(e)
#             install('rasterio')
#             import rasterio as rio

#         try:
#             import geopandas as gpd
#         except Exception as e:
#             print(e)
#             install('geopandas')
#             import geopandas as gpd
#         try:
#             import xarray as xr
#         except Exception as e:
#             print(e)
#             install('xarray')
#             import xarray as xr
#         try:
#             import rioxarray as rxr
#         except Exception as e:
#             print(e)
#             install('rioxarray')
#             import rioxarray as rxr

#     def plot_load():
#         global sns
#         try:
#             import seaborn as sns
#         except Exception as e:
#             print(e)
#             install('seaborn')
#             import seaborn as sns

#     def miscellaneous_load():
#         global pip, imod
#         import pip
#         from operator import imod

#     if opt == 0:
#         pass
#     elif opt == 1:
#         geo_load()
#     elif opt == 2:
#         geo_load()
#         plot_load()
#     elif opt == 3:
#         geo_load()
#         plot_load()
#         miscellaneous_load()
#     else:
#         raise  Exception(
#             '''
#             opt should be one of the following: 
#             >>> 0: no action (default)
#             >>> 1: loading rasterio, geopandas, xarray, and rioxarray
#             >>> 2: 1 + seaborn
#             >>> 3: 2 + pip and imod
#             '''
#         )
#     print('Selected packages are installed...')


colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']



import numpy as np

# # Define the Arrhenius equation for respiration
# def arrhenius_temperature_dependence(T, E_a, R=8.314, T_0=293.15):
#     """
#     Calculate temperature-dependent respiration using Arrhenius equation.
#     T: temperature (in Kelvin)
#     E_a: activation energy (J/mol)
#     R: universal gas constant (J/mol·K)
#     T_0: reference temperature (in Kelvin)
#     """
#     return np.exp(E_a / R * (1 / T - 1 / T_0))

# # Heterotrophic respiration function (simple version)
# def heterotrophic_respiration(T, E_h, moisture_factor=1, R=8.314, T_0=293.15):
#     """
#     Calculate heterotrophic respiration using temperature dependence and moisture factor.
#     T: temperature in Celsius
#     E_h: activation energy for heterotrophic respiration (J/mol)
#     moisture_factor: scaling factor for moisture (default = 1, can be adjusted)
#     """
#     T_K = T + 273.15  # Convert to Kelvin
#     return arrhenius_temperature_dependence(T_K, E_h, R, T_0) * moisture_factor

# # Example function to calculate autotrophic respiration
# def calculate_autotrophic_respiration(T, E_a=60000):
#     """
#     Calculate autotrophic respiration from temperature.
#     T: temperature in Celsius
#     E_a: activation energy for autotrophic respiration (J/mol)
#     """
#     T_K = T + 273.15  # Convert to Kelvin
#     return arrhenius_temperature_dependence(T_K, E_a)

# def calculate_ecosystem_respiration_obsolete(T, E_a=60000, E_h=60000, moisture_factor = 1, R = 8.314, T_0 = 293.15):
#     """
#     Calculate the total ecosystem respiration using autotrophic and heterotrophic respiration.
#     T: temperature in Celsius
#     E_a: activation energy for autotrophic respiration (J/mol)
#     E_h: activation energy for heterotrophic respiration (J/mol)
#     moisture_factor: scaling factor for moisture (for heterotrophic, default = 1, can be adjusted) 
#     R: universal gas constant (J/mol·K)
#     T_0: reference temperature (in Kelvin)

#     Ea = 0.65 eV = ca.62.4 kJ/mol
#     Biome-scale temperature sensitivity of ecosystem respiration revealed by atmospheric CO2 observations
#     """
#     Ra = calculate_autotrophic_respiration(T, E_a)
#     Rh = heterotrophic_respiration(T, E_h, moisture_factor, R, T_0)
#     Reco = Ra + Rh
    
#     return Reco


def calculate_ecosystem_respiration(Ta, Rref, Q10=1.5, Tref=10):
    """
    Calculate ecosystem respiration using the Q10 model.

    Parameters:
    - Ta (float or array): Ambient temperature (degC).
    - Rref (float): Reference respiration rate at the reference temperature (Tref) [e.g., umol CO2 m-2 s-1].
    - Q10 (float, optional): Temperature sensitivity factor (default is 1.5).
    - Tref (float, optional): Reference temperature (degC, default is 10).

    Returns:
    - float or array: Ecosystem respiration (Reco) at temperature Ta.
    """
    B = np.log(Q10) / 10  # Temperature sensitivity scaling factor
    Reco = Rref * np.exp(B * (Ta - Tref))  # Ecosystem respiration
    return Reco


import numpy as np

def get_NDVI(r, nir):
    ndvi = (nir - r) / (nir + r)
    return ndvi

def get_NIRv(ndvi, nir):
    nirv = ndvi * nir
    return nirv

def get_kNDVI(ndvi):
    kndvi = np.tanh(ndvi**2)
    return kndvi

def get_EVI2band(r, nir):
    evi = 2.5 * (nir - r) / (nir + 2.4 * r + 1)
    return evi


def get_MODIS_IGBPcode():
    '''
    url: https://developers.google.com/earth-engine/datasets/catalog/MODIS_061_MCD12Q1 
    '''
    MODIS_IGBP_codes = ['ENF', 'EBF', 'DNF', 'DBF', 'MF', 'CSH', 'OSH', 'WSA', 'SAV', 'GRA', 'WET', 'CRO', 'URB', 'CNV', 'SNO', 'BSV', 'WAB']
    MODIS_IGBP_dict = dict(zip(MODIS_IGBP_codes, np.arange(len(MODIS_IGBP_codes)) + 1))
    return MODIS_IGBP_dict

def convert_gCm2d1_PgCyr_025deg():
    '''
    Terrestrial ecosystem carbon flux unit conversion:
    from gC m-2 d-1 to Pg C yr-1 for spatial resolution of 0.25 deg
    '''
    coef = 365 * 0.25 * 0.25 * 1e5 * 1e5 / 1e15
    return coef

def deg2m(longitude, latitude, scale_lon, scale_lat):
    # deg x deg => m2
    # Length in km of 1° of latitude = always 111.32 km
    # Length in km of 1° of longitude = 40075 km * cos( latitude ) / 360
    _, lats = np.meshgrid(longitude, latitude)
    coef_mat = 40075 * np.cos(np.deg2rad(np.abs(lats))) / 360 * 111.32 * 1e3 * 1e3
    coef_mat = coef_mat * scale_lon * scale_lat
    return coef_mat


import numpy as np

'''
Source: https://github.com/ImperialCollegeLondon/pyrealm
'''

def calc_co2_to_ca(co2, patm):
    r"""CO2 ppm to Pa"""

    return 1.0e-6 * co2 * patm  # Pa, atms. CO2

def calc_ftemp_arrh(tk, ha):
    """Calculate enzyme kinetics scaling factor."""
    # Standard baseline reference temperature of photosynthetic processes (Prentice,
    # unpublished)  (:math:`T_o` , 25.0, °C)
    plant_T_ref = 25.0
    # Conversion from °C to K   (:math:`CtoK` , 273.15, -)
    k_CtoK = 273.15
    # Universal gas constant (:math:`R` , 8.3145, J/mol/K)
    k_R = 8.3145

    tkref = plant_T_ref + k_CtoK

    return np.exp(ha * (tk - tkref) / (tkref * k_R * tk))


def calc_gammastar(tc, patm):
    # Calculate the photorespiratory CO2 compensation point.
    # Bernacchi estimate of gs25_0
    bernacchi_gs25_0 = 4.332  # Reported as 42.75 µmol mol-1
    # Standard reference atmosphere (Allen, 1973) (:math:`P_o` , 101325.0, Pa)
    k_Po = 101325.0
    # Conversion from °C to K   (:math:`CtoK` , 273.15, -)
    k_CtoK = 273.15
    # Bernacchi estimate of activation energy for gammastar (J/mol)
    bernacchi_dha = 37830
    return (
        bernacchi_gs25_0 * patm / k_Po
        * calc_ftemp_arrh((tc + k_CtoK), ha = bernacchi_dha)
    )

def evaluate_horner_polynomial(x, cf):
    """Evaluates a polynomial with coefficients `cf` at `x` using Horner's method."""
    y = np.zeros_like(x)
    for c in reversed(cf):
        y = x * y + c
    return y

def calc_density_h2o_fisher(tc, patm):
    """Calculate water density."""

    # Calculate lambda, (bar cm^3)/g:
    # Fisher Dial
    fisher_dial_lambda = np.array([1788.316, 21.55053, -0.4695911, 0.003096363, -7.341182e-06])

    lambda_coef = fisher_dial_lambda
    lambda_val = evaluate_horner_polynomial(tc, lambda_coef)

    # Calculate po, bar
    fisher_dial_Po = np.array([5918.499, 58.05267, -1.1253317, 0.0066123869, -1.4661625e-05])
    po_coef = fisher_dial_Po
    po_val = evaluate_horner_polynomial(tc, po_coef)

    # Calculate vinf, cm^3/g

    fisher_dial_Vinf = np.array([
        0.6980547, -0.0007435626, 3.704258e-05, -6.315724e-07, 9.829576e-09,
        -1.197269e-10, 1.005461e-12, -5.437898e-15, 1.69946e-17, -2.295063e-20
    ])

    vinf_coef = fisher_dial_Vinf
    vinf_val = evaluate_horner_polynomial(tc, vinf_coef)

    # Convert pressure to bars (1 bar <- 100000 Pa)
    pbar = 1e-5 * patm

    # Calculate the specific volume (cm^3 g^-1):
    spec_vol = vinf_val + lambda_val / (po_val + pbar)

    # Convert to density (g cm^-3) -> 1000 g/kg; 1000000 cm^3/m^3 -> kg/m^3:
    rho = 1e3 / spec_vol
    return rho

def calc_density_h2o_chen(tc, p):
    """Calculate the density of water using Chen et al 2008."""

    # Calculate density at 1 atm (kg/m^3):
    # Chen water density
    chen_po = np.array([
        0.99983952, 6.788260e-5, -9.08659e-6, 1.022130e-7, -1.35439e-9,
        1.471150e-11, -1.11663e-13, 5.044070e-16, -1.00659e-18,
    ])
    po_coef = chen_po
    po = evaluate_horner_polynomial(tc, po_coef)

    # Calculate bulk modulus at 1 atm (bar):
    chen_ko = np.array([19652.17, 148.1830, -2.29995, 0.01281, -4.91564e-5, 1.035530e-7])
    ko_coef = chen_ko
    ko = evaluate_horner_polynomial(tc, ko_coef)

    # Calculate temperature dependent coefficients:
    chen_ca = np.array([3.26138, 5.223e-4, 1.324e-4, -7.655e-7, 8.584e-10])
    ca_coef = chen_ca
    ca = evaluate_horner_polynomial(tc, ca_coef)

    chen_cb = np.array([7.2061e-5, -5.8948e-6, 8.69900e-8, -1.0100e-9, 4.3220e-12])
    cb_coef = chen_cb
    cb = evaluate_horner_polynomial(tc, cb_coef)

    # Convert atmospheric pressure to bar (1 bar = 100000 Pa)
    pbar = (1.0e-5) * p

    pw = ko + ca * pbar + cb * pbar**2.0
    pw /= ko + ca * pbar + cb * pbar**2.0 - pbar
    pw *= (1e3) * po
    return pw

def calc_density_h2o(tc, patm, safe = True, water_density_method = 'fisher'):
    """Calculate water density."""

    # Safe guard against instability in functions at low temperature.
    if safe and np.nanmin(tc) < np.array([-30]):
        raise ValueError(
            "Water density calculations below about -30°C are "
            "unstable. See argument safe to calc_density_h2o"
        )

    if water_density_method == "fisher":
        return calc_density_h2o_fisher(tc, patm)

    if water_density_method == "chen":
        return calc_density_h2o_chen(tc, patm)

    raise ValueError("Unknown method provided to calc_density_h2o")

def calc_viscosity_h2o(tc, patm, simple = False, simple_viscosity = False):
    """Calculate the viscosity of water."""

    if simple or simple_viscosity:
        # The reference for this is unknown, but is used in some implementations
        # so is included here to allow intercomparison.
        return np.exp(-3.719 + 580 / ((tc + 273) - 138))

    # Get the density of water, kg/m^3
    rho = calc_density_h2o(tc, patm)

    # Calculate dimensionless parameters:
    # Conversion from °C to K   (:math:`CtoK` , 273.15, -)
    k_CtoK = 273.15
    # Huber reference temperature (:math:`tk_{ast}`, 647.096, Kelvin)
    huber_tk_ast = 647.096
    # Huber reference density (:math:`\rho_{ast}`, 322.0, kg/m^3)
    huber_rho_ast = 322.0
    tbar = (tc + k_CtoK) / huber_tk_ast
    rbar = rho / huber_rho_ast

    # Calculate mu0 (Eq. 11 & Table 2, Huber et al., 2009):
    # Temperature dependent parameterisation of Hi in Huber.
    huber_H_i = np.array([1.67752, 2.20462, 0.6366564, -0.241605])
    mu0 = huber_H_i[0] + huber_H_i[1] / tbar
    mu0 += huber_H_i[2] / (tbar * tbar)
    mu0 += huber_H_i[3] / (tbar * tbar * tbar)
    mu0 = (1e2 * np.sqrt(tbar)) / mu0

    # Calculate mu1 (Eq. 12 & Table 3, Huber et al., 2009):
    ctbar = (1.0 / tbar) - 1.0
    mu1 = 0.0

    # Iterate over the rows of the H_ij core_constants matrix
    # Temperature and mass density dependent parameterisation of Hij in Huber.
    huber_H_ij = np.array([
                [0.520094, 0.0850895, -1.08374, -0.289555, 0.0, 0.0],
                [0.222531, 0.999115, 1.88797, 1.26613, 0.0, 0.120573],
                [-0.281378, -0.906851, -0.772479, -0.489837, -0.25704, 0.0],
                [0.161913, 0.257399, 0.0, 0.0, 0.0, 0.0],
                [-0.0325372, 0.0, 0.0, 0.0698452, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.00872102, 0.0],
                [0.0, 0.0, 0.0, -0.00435673, 0.0, -0.000593264],
            ])

    for row_idx in np.arange(huber_H_ij.shape[1]):
        cf1 = ctbar**row_idx
        cf2 = 0.0
        for col_idx in np.arange(huber_H_ij.shape[0]):
            cf2 += huber_H_ij[col_idx, row_idx] * (rbar - 1.0) ** col_idx
        mu1 += cf1 * cf2

    mu1 = np.exp(rbar * mu1)

    # Calculate mu_bar (Eq. 2, Huber et al., 2009), assumes mu2 = 1
    mu_bar = mu0 * mu1

    # Calculate mu (Eq. 1, Huber et al., 2009)
    # Huber reference pressure (:math:`\mu_{ast}` 1.0e-6, Pa s)
    huber_mu_ast = 1e-06
    return mu_bar * huber_mu_ast  # Pa s

def calc_ns_star(tc, patm):
    r"""Calculate the relative viscosity of water."""

    visc_env = calc_viscosity_h2o(tc, patm)
    # Standard reference temperature (:math:`T_o` ,  298.15, K)
    k_To = 298.15
    # Conversion from °C to K   (:math:`CtoK` , 273.15, -)
    k_CtoK = 273.15
    # Standard reference atmosphere (Allen, 1973) (:math:`P_o` , 101325.0, Pa)
    k_Po = 101325.0
    visc_std = calc_viscosity_h2o(np.array(k_To) - np.array(k_CtoK), np.array(k_Po))

    return visc_env / visc_std

def calc_kmm(tc, patm):
    """Calculate the Michaelis Menten coefficient of Rubisco-limited assimilation."""

    # conversion to Kelvin
    # Conversion from °C to K   (:math:`CtoK` , 273.15, -)
    k_CtoK = 273.15
    tk = tc + k_CtoK

    # Bernacchi estimate of kc25
    bernacchi_kc25 = 39.97  # Reported as 404.9 µmol mol-1
    # Bernacchi estimate of activation energy Kc for CO2 (J/mol)
    bernacchi_dhac = 79430
    kc = bernacchi_kc25 * calc_ftemp_arrh(
        tk, ha = bernacchi_dhac
    )
    # Bernacchi estimate of ko25
    bernacchi_ko25 = 27480  # Reported as 278.4 mmol mol-1
    # Bernacchi estimate of activation energy Ko for O2 (J/mol)
    bernacchi_dhao = 36380
    ko = bernacchi_ko25 * calc_ftemp_arrh(
        tk, ha = bernacchi_dhao
    )

    # O2 partial pressure
    # O2 partial pressure, Standard Atmosphere (:math:`co` , 209476.0, ppm)
    k_co = 209476.0
    po = k_co * 1e-6 * patm

    return kc * (1.0 + po / ko)


def wang17(mj):
    """Calculate limitation factors following :cite:`Wang:2017go`."""
    # Unit carbon cost for the maintenance of electron transport capacity (:math:`c`, 0.41, )
    wang17_c = 0.41

    vals_defined = np.greater(mj, wang17_c)

    f_v = np.sqrt(1 - (wang17_c / mj) ** (2.0 / 3.0), where=vals_defined)
    f_j = np.sqrt((mj / wang17_c) ** (2.0 / 3.0) - 1, where=vals_defined)

    # Backfill undefined values - tackling float vs np.ndarray types
    if isinstance(f_v, np.ndarray):
        f_j[np.logical_not(vals_defined)] = np.nan  # type: ignore
        f_v[np.logical_not(vals_defined)] = np.nan  # type: ignore
    elif not vals_defined:
        f_j = np.nan
        f_v = np.nan
    return f_j, f_v

def smith19(mj):
    """Calculate limitation factors following :cite:`Smith:2019dv`."""

    # Adopted from Nick Smith's code:
    # Calculate omega, see Smith et al., 2019 Ecology Letters  # Eq. S4
    # Scaling factor theta for Jmax limitation (:math:`\theta`, 0.85)
    smith19_theta = 0.85
    theta = smith19_theta
    # Scaling factor c for Jmax limitation (:math:`c`, 0.05336251)
    smith19_c_cost = 0.05336251
    c_cost = smith19_c_cost

    # simplification terms for omega calculation
    cm = 4 * c_cost / mj
    v = 1 / (cm * (1 - smith19_theta * cm)) - 4 * theta

    # account for non-linearities at low m values. This code finds
    # the roots of a quadratic function that is defined purely from
    # the scalar theta, so will always be a scalar. The first root
    # is then used to set a filter for calculating omega.

    cap_p = (((1 / 1.4) - 0.7) ** 2 / (1 - theta)) + 3.4
    aquad = -1
    bquad = cap_p
    cquad = -(cap_p * theta)
    roots = np.polynomial.polynomial.polyroots(
        [aquad, bquad, cquad]
    )  # type: ignore [no-untyped-call]

    # factors derived as in Smith et al., 2019
    m_star = (4 * c_cost) / roots[0].real
    omega = np.where(
        mj < m_star,
        -(1 - (2 * theta)) - np.sqrt((1 - theta) * v),
        -(1 - (2 * theta)) + np.sqrt((1 - theta) * v),
    )

    # np.where _always_ returns an array, so catch scalars
    omega = omega.item() if np.ndim(omega) == 0 else omega

    omega_star = (
        1.0
        + omega
        - np.sqrt((1.0 + omega) ** 2 - (4.0 * theta * omega))  # Eq. 18
    )

    # Effect of Jmax limitation - note scaling here. Smith et al use
    # phi0 as as the quantum efficiency of electron transport, which is
    # 4 times our definition of phio0 as the quantum efficiency of photosynthesis.
    # So omega*/8 theta and omega / 4 are scaled down here  by a factor of 4.
    # Ignore `mypy` here as omega_star is explicitly not None.
    f_v = omega_star / (2.0 * theta)  # type: ignore
    f_j = omega
    return f_j, f_v

def simple():
    """Apply the 'simple' form of the equations."""

    # Set Jmax limitation to unity - could define as 1.0 in __init__ and
    # pass here, but setting explicitly within the method for clarity.
    f_v = np.array([1.0])
    f_j = np.array([1.0])
    return f_j, f_v

def calc_ftemp_kphio(tc, c4 = False):
    # Quadratic scaling of Kphio with temperature
    kphio_C4 = (-0.064, 0.03, -0.000464)
    kphio_C3 = (0.352, 0.022, -0.00034)

    if c4:
        coef = kphio_C4
    else:
        coef = kphio_C3

    ftemp = coef[0] + coef[1] * tc + coef[2] * tc**2
    ftemp = np.clip(ftemp, 0.0, None)

    return ftemp


def calc_light_water_use_efficiency(tc, patm, ca, vpd, do_ftemp_kphio, c4 = False, limitation_factors = 'wang17'):
    """
    The basic calculation of LUE = phi0 * M_c * m with an added penalty term for jmax limitation
    """
    k_c_molmass = 12.0107
    # Molecular mass of carbon (:math:`c_molmass` , 12.0107, g)

    # Set context specific defaults for kphio to match Stocker paper
    if not do_ftemp_kphio:
        init_kphio = 0.049977
    else:
        init_kphio = 0.081785

    if do_ftemp_kphio:
        ftemp_kphio = calc_ftemp_kphio(tc, c4 = c4)
        kphio = init_kphio * ftemp_kphio
    else:
        kphio = np.array([init_kphio])

    gammastar = calc_gammastar(tc, patm)
    ns_star = calc_ns_star(tc, patm)
    kmm = calc_kmm(tc, patm)

    if c4:
        # Unit cost ratio for C4 plants (:math:`\beta`, 16.222).
        beta_cost_ratio_c4 = np.array([146.0 / 9])
        beta = beta_cost_ratio_c4

    else:
        beta_cost_ratio_prentice14 = np.array([146.0])
        beta = beta_cost_ratio_prentice14

    xi = np.sqrt((beta * (kmm + gammastar)) / (1.6 * ns_star))

    chi = gammastar / ca + (1.0 - gammastar / ca) * xi / (xi + np.sqrt(vpd))

    ci = chi * ca
    if c4:
        mj = 1
    else:
        mj = (ci - gammastar) / (ci + 2 * gammastar)

    if limitation_factors == 'wang17':
        f_j, f_v = wang17(mj)
    elif limitation_factors == 'smith19':
        f_j, f_v = smith19(mj)
    else:
        f_j, f_v = simple()

    lue = kphio * mj * f_v * k_c_molmass
    # Intrinsic water use efficiency (iWUE, µmol mol-1)
    iwue = (5 / 8 * (ca - ci)) / (1e-6 * patm)

    return lue, iwue

'''
# Example
# fapar: the fraction of absorbed photosynthetically active radiation (-)
# ppfd: photosynthetic photon flux density (µmol m-2 s-1)
# iabs = fapar * ppfd

tc = 20
patm = 101325.0
co2 = 400
vpd = 100
do_ftemp_kphio = True

ca = calc_co2_to_ca(co2, patm)
# # C3:
print(
    calc_light_water_use_efficiency(tc, patm, ca, vpd, do_ftemp_kphio, c4 = False, limitation_factors = 'wang17')
)
# # C4:
print(
    calc_light_water_use_efficiency(tc, patm, ca, vpd, do_ftemp_kphio, c4 = True, limitation_factors = 'none')
)

# -----------------------------
# Check against

# !pip install pyrealm

from pyrealm import pmodel

env  = pmodel.PModelEnvironment(tc=tc, patm=patm, vpd=vpd, co2=co2)
# # C3
print(
    pmodel.PModel(env, method_optchi='prentice14').lue
)
# # C4
print(
    pmodel.PModel(env, method_optchi='c4', method_jmaxlim='none').lue
)

# mod_c4 = PModel(env, method_optchi='c4', method_jmaxlim='none')
'''

# env  = pmodel.PModelEnvironment(tc=dfp['TA_F_MDS'].values, patm=dfp['PA'].values * 100, vpd=dfp['VPD_F_MDS'].values * 100, co2=dfp['CO2_F_MDS'].values)
# model = pmodel.PModel(env, method_optchi=c3c4) # or c4
# model.estimate_productivity(fapar=dfp['NIRv'].values * 0.5, ppfd=dfp['PPFD_IN'].values)


from . import easy_import
from . import google
from . import stutils
from . import sciplt
from . import bigleaf
from . import commons
# from .regress2 import regress2
# from .lins_ccc_1989 import concordance_correlation_coefficient
# from .manage_yaml import Yaml
# from .manage_file import create_all_parents, searching_all_files, unzip, pbar

__all__ = [
    "easy_import",
    "google",
    "stutils",
    "sciplt",
    "bigleaf",
    "commons",
    ]


import pandas as pd

def get_ONI():
    '''
    El Niño and La Niña Years and Intensities
    Based on Oceanic Niño Index (ONI)
    https://ggweather.com/enso/oni.htm
    
    Covered periods: 2000 - 2023
    '''
    dict_el = {
        1 : [2004, 2006, 2014, 2018], # 'El-Nino-weak'
        2: [2002, 2009], # 'El-Nino-moderate'
        3: [2023], # 'El-Nino-strong'
        4: [2015], # 'El-Nino-very-strong'
        0: [2001, 2003, 2012, 2013, 2019], # None
        -1: [2000, 2005, 2008, 2016, 2017], # 'La-Nina-weak'
        -2: [2011, 2020, 2021, 2022], # 'La-Nina-moderate'
        -3: [2007, 2010], #  'La-Nina-strong'
    }

    years_el = []; values_el = []
    for k, v in dict_el.items():
        values_el.extend(v)
        for i in v:
            years_el.append(k)

    df_el = pd.DataFrame({'El-Nino-La-Nina': years_el}, index = values_el).sort_index()

    return df_el


# sort path by time
path_time = [pd.to_datetime("".join(p.stem.split("_")[-4::]), format = "%Y%m%d%H%M") for p in paths]
# [x for _, x in sorted(zip(Y, X))] # sort X by Y values
paths = [p for _, p in sorted(zip(path_time, paths))]
# ========================================================================

if float(lat) >= 0:
    dfc["seasons"] = (dfc.index.month%12 + 3)//3
else:
    dfc["seasons"] = ((dfc.index.month + 6)%12 + 3)//3
# ========================================================================
x = 3.4136
print(f"{x}: .2f")

# ========================================================================
# boxplot

import seaborn as sns

air_poll_names = ['$AOD_{550}$', '$O_3$', '$NO_2$', '$HCHO$']
shap_names = [name + " (SHAP)" for name in air_poll_names]
print(shap_values[:, '$O_3$'].shape, X.shape)
# print(shap_values[:, '$O_3$'].data, X['$O_3$'])
df_analysis = X[air_poll_names].copy()
df_analysis["Season"] = seasons
df_analysis["IGBP"] = igbp_series
df_analysis["ID"] = site_series
for name in air_poll_names:
    df_analysis[name + " (SHAP)"] = shap_values[:, name].values
# print(df_analysis.columns)
dfp = df_analysis.copy()

# site average
dfp = df_analysis.groupby(['ID']).median()
meta = df_analysis[["ID", "IGBP"]].copy()
meta = meta.drop_duplicates(subset='ID', keep="last")
meta = meta.set_index("ID", drop = True)
dfp = pd.concat([dfp, meta], axis = 1)
dfp = dfp.sort_values(by = "IGBP")

# # use pandas to plot manually
# dfp = dfp[[air_poll_names[0], "IGBP"]]
# dfp = pd.pivot_table(dfp, values = air_poll_names[0], index = dfp.index, columns=['IGBP']).reset_index()
# dfp.boxplot()

# use seaborn
dfp = pd.melt(dfp, id_vars=['IGBP'], value_vars=shap_names,
        var_name='Pollutant', value_name="$SHAP \; (gC \; m^{-2} \; d^{-1})$"
        )

# sns.set_style("white")
sns.set(rc={'figure.figsize':(12, 10)}, style = "white", font_scale = 1.5)
sns.set_style("ticks", {"xtick.major.size": 8, "ytick.major.size": 8})
ax = sns.boxplot(x="IGBP", y="$SHAP \; (gC \; m^{-2} \; d^{-1})$", hue="Pollutant",
                 data=dfp, palette="Set3")
ax.axhline(y=0.0, color='k', linestyle='--', alpha = 0.5)
ax.set_ylim(-0.3, 0.3)
plt.tick_params(direction = "in", which = "both")
fig = ax.get_figure()

# Python > 3.7
handles, labels = ax.get_legend_handles_labels()
by_label = dict(zip(labels, handles))
ax.legend(by_label.values(), by_label.keys(), loc = "upper center", framealpha = 0.1, frameon = True , bbox_to_anchor=(0.5, 1.1), ncol = 4)

# ============================================================================
# split one column into three and rename them
df = df.join(df["date"].str.split("_", expand=True).astype(int).rename(columns = {0: "year", 1: "month", 2: "day"}))

# ============================================================================
# vertical barplots

# site wise r2

import seaborn as sns
import matplotlib.pyplot as plt
sns.set(style="whitegrid")

df = read_df_dn()
df["site"] = df["site"] + "(" + df["igbp"] + ")"

time_type = "diel"

# print(df.columns)

df["total"] = df[f"MDS_{time_type}_r2"] + df[f"RFR_{time_type}_r2"]
df["dif_r2"] = df[f"RFR_{time_type}_r2"] - df[f"MDS_{time_type}_r2"]
# Initialize the matplotlib figure
f, ax = plt.subplots(figsize=(8, 18))

df = df.sort_values("total", ascending=False)
# split fig
df = df.iloc[100::, :]
print(len(df))

# Plot the total crashes
sns.set_color_codes("pastel")
g = sns.barplot(x="total", y="site", data=df,
            label=f"MDS $r^2$", color="b")

# Plot the crashes where alcohol was involved
sns.set_color_codes("muted")
sns.barplot(x=f"RFR_{time_type}_r2", y="site", data=df,
            label=f"RFR $r^2$", color="b")

# Add a legend and informative axis label
ax.legend(ncol=2, loc="lower right", frameon=True)
ax.set(xlim=(0, 2), ylabel="",
       xlabel=f"$The \; r^2 \; comparison$")
sns.despine(left=True, bottom=True)

for i in range(len(df)):
    p = ax.patches[i]
    height = p.get_height()
    width = p.get_width()
    if p.get_x() + width >= 8:
        x_pos = 8
    else:
        x_pos = p.get_x() + width
    # ax.text(x_pos, p.get_y() + height/2, np.round(df["dif_bias"].values[i], 2))
    ax.annotate(f"{np.round(df['dif_r2'].values[i], 2)}", (x_pos, p.get_y() + height))
    

f.savefig(f"3drivers/site_r2_compare_sort_dif_{time_type}_b.pdf", dpi = 600, format = "pdf", bbox_inches = "tight")

# ========================================================================================================================
# drop rows with duplicated index
df = df[~df.index.duplicated(keep='first')]
df = df.loc[:,~df.columns.duplicated()]

# ========================================================================================================================
# x-axis datetime format
import matplotlib.dates as mdates
dtfmt = mdates.DateFormatter('%d/%m/%y %H:%m')
for ax in axes:
    ax.xaxis.set_major_formatter(dtfmt)
# =========================================================================================================================
# Diurnal plots

for df_orig, site_name in zip([seb, sab], ["Sebungan", "Sabaju"]):
    df_orig = df_orig.replace(-9999., np.nan)
    for method in df_orig.columns:

        df = df_orig[[method]].copy()
        df['Time'] = df.index.map(lambda x: x.strftime("%H:%M"))
        df['Month'] = df.index.map(lambda x: x.strftime("%m"))
        df['Year'] = df.index.map(lambda x: x.strftime("%Y"))
        
        plot_diurnal(df, method, site_name)
        
print("done")

import matplotlib.pyplot as plt

def plot_diurnal(df, method, site_name):
    df_diurnal = df.pivot_table(method, ['Year', 'Month', 'Time'], aggfunc='mean').reset_index()
    months = list(df_diurnal["Month"].unique())
    months.sort(key=float)

    fig, axs = plt.subplots(figsize=(16, 12), 
                            nrows = 4, ncols = 3,
                            sharex = True,
                            sharey = True
                           )

    for mt, ax in zip(months, axs.flatten()):
        ax.tick_params(direction = "in")
        mt_label = datetime.strptime(mt, "%m").strftime("%B")
        dft = df_diurnal.query("Month == @mt")
        # years = dft["Year"].unique()
        dft = dft.drop("Month", axis = 1)
        dft = dft.set_index("Time")
        # print(dft)
        # # ==================================================================
        # # method 1: plot lines in one subfigure automatedly but has issues with legend
        # dft.groupby("Year")[method].plot(legend=True, ax=ax, title=mt_label)
        
        # # ==================================================================
        # # method 2: plot lines and legend manually
        years = [2016, 2017, 2018, 2019, 2020]
        colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd']
        legends = [str(lg) for lg in years]
        for yr, c in zip(years, colors):
            if dft[dft["Year"] == str(yr)].empty: continue
            dft[dft["Year"] == str(yr)].plot(color = c, ax = ax, legend = None, title=mt_label)
            # lgds.append(str(yr))
        # ax.legend(lgds)
        ax.set_ylabel("NEE $(\mu mol \; m^{-2} \; d^{-1})$")
        ax.set_xlabel("Hour", fontsize = 12)
    #     break
    
    # =========================================================================================================
    # extract common legends
    lines = []
    labels = []

    ax_colors = []
    for ax in fig.axes:
        axLine, axLabel = ax.get_legend_handles_labels()
        for line2D in axLine:
            if line2D.get_color() in ax_colors:
                continue
            else:
                ax_colors.append(line2D.get_color())
                line2D.set_label(
                    dict(zip(colors, legends))[line2D.get_color()]
                )
                lines.append(line2D)
    lines = sorted(lines, key=lambda x: int(x.get_label()), reverse=False)
    axs.flatten()[-2].legend(handles = lines, loc='upper center', bbox_to_anchor=(0.5, -0.2), ncol=len(lines))
    # =========================================================================================================
    plt.suptitle(method.upper(), fontsize = 16, x = 0.15, y = 0.93)

    plt.savefig(f"{site_name}-{method}-diurnal.png", bbox_inches = "tight", dpi = 300)
    
    plt.close(fig)
# =========================================================================================================================
# increase marker size in legend
legend = ax.legend(loc = "lower left")
for handle in legend.legendHandles:
    handle.set_sizes([20])
# =========================================================================================================================
# tick rotation
ax.tick_params(axis='x', rotation=45)
# =========================================================================================================================
# Efficient way to group indices of the same elements in a list
data = [1, 2, 2, 5, 8, 3, 3, 9, 0, 1]
pd.Series(range(len(data))).groupby(data, sort=False).apply(list).tolist()
# [[0, 9], [1, 2], [3], [4], [5, 6], [7], [8]]

# =========================================================================================================================
# Intersection of multiple lists
from functools import reduce
# d = [list1, list2, ...]
# solution1:
set.intersection(*map(set,d))
# solution2:
# apply intersect1d to (a list of) multiple lists:
reduce(np.intersect1d, d)


# =========================================================================================================================
'''
Cirular bar plot
'''
# https://www.python-graph-gallery.com/circular-barplot-with-groups

OFFSET = np.pi / 2

def get_label_rotation(angle, offset):
    # Rotation must be specified in degrees :(
    rotation = np.rad2deg(angle + offset)
    if angle <= np.pi:
        alignment = "right"
        rotation = rotation + 180
    else: 
        alignment = "left"
    return rotation, alignment

def add_labels(angles, values, labels, offset, ax):
    
    # This is the space between the end of the bar and the label
    padding = 4
    
    # Iterate over angles, values, and labels, to add all of them.
    for angle, value, label, in zip(angles, values, labels):
        angle = angle
        
        # Obtain text rotation and alignment
        rotation, alignment = get_label_rotation(angle, offset)

        # And finally add the text
        ax.text(
            x=angle, 
            y=value + padding, 
            s=label, 
            ha=alignment, 
            va="center", 
            rotation=rotation, 
            rotation_mode="anchor"
        ) 


# Ensures reproducibility of random numbers
rng = np.random.default_rng(123)

# Build a dataset
df = pd.DataFrame({
    "name": [f"item {i}" for i in range(1, 51)],
    "value": rng.integers(low=30, high=100, size=50),
    "group": ["A"] * 10 + ["B"] * 20 + ["C"] * 12 + ["D"] * 8
})

# # Build a dataset
# df = pd.DataFrame({
#     "name": [f"item {i}" for i in range(1, 32 * 6 + 1)],
#     "value": rng.integers(low=30, high=100, size=32 * 6),
#     "group": ["A"] * 32 + ["B"] * 64 + ["C"] * 32 + ["D"] * 64
# })

# Show 3 first rows
df.head(3)


# All this part is like the code above
VALUES = df["value"].values
LABELS = df["name"].values
GROUP = df["group"].values

PAD = 3
ANGLES_N = len(VALUES) + PAD * len(np.unique(GROUP))
ANGLES = np.linspace(0, 2 * np.pi, num=ANGLES_N, endpoint=False)
WIDTH = (2 * np.pi) / len(ANGLES)

offset = 0
IDXS = []
GROUPS_SIZE = [10, 20, 12, 8]
# GROUPS_SIZE = [32, 64, 32, 64]
for size in GROUPS_SIZE:
    IDXS += list(range(offset + PAD, offset + size + PAD))
    offset += size + PAD

fig, ax = plt.subplots(figsize=(20, 10), subplot_kw={"projection": "polar"})
ax.set_theta_offset(OFFSET)
ax.set_ylim(-100, 100)
ax.set_frame_on(False)
ax.xaxis.grid(False)
ax.yaxis.grid(False)
ax.set_xticks([])
ax.set_yticks([])

# GROUPS_SIZE = [10, 20, 12, 8]
COLORS = [f"C{i}" for i, size in enumerate(GROUPS_SIZE) for _ in range(size)]

ax.bar(
    ANGLES[IDXS], VALUES, width=WIDTH, color=COLORS, 
    edgecolor="white", linewidth=2
)

add_labels(ANGLES[IDXS], VALUES, LABELS, OFFSET, ax)

# Extra customization below here --------------------

# This iterates over the sizes of the groups adding reference
# lines and annotations.

offset = 0 
for group, size in zip(["A", "B", "C", "D"], GROUPS_SIZE):
    # Add line below bars
    x1 = np.linspace(ANGLES[offset + PAD], ANGLES[offset + size + PAD - 1], num=50)
    ax.plot(x1, [-5] * 50, color="#333333")
    
    # Add text to indicate group
    ax.text(
        np.mean(x1), -20, group, color="#333333", fontsize=14, 
        fontweight="bold", ha="center", va="center"
    )
    
    # Add reference lines at 20, 40, 60, and 80
    x2 = np.linspace(ANGLES[offset], ANGLES[offset + PAD - 1], num=50)
    ax.plot(x2, [20] * 50, color="#bebebe", lw=0.8)
    ax.plot(x2, [40] * 50, color="#bebebe", lw=0.8)
    ax.plot(x2, [60] * 50, color="#bebebe", lw=0.8)
    ax.plot(x2, [80] * 50, color="#bebebe", lw=0.8)
    
    offset += size + PAD


# =========================================================================================================================
# Transform data to axis coordinates:

ax.transData.transform((2, 15))


# =========================================================================================================================
# element-wise add two arrays ignoring NaNs
# https://stackoverflow.com/questions/33269369/adding-two-2d-numpy-arrays-ignoring-nans-in-them
np.nansum(np.dstack((A,B)),2)


# =========================================================================================================================
# Offset time
df02.index = df02.index + pd.tseries.offsets.Minute(1)
pd.DateOffset(months=1)

# ==========================================================================================================================
# list comprehension if else
[x+1 if x >= 45 else x+5 for x in l]

# ========================================================================================================================================================
# matplotlib default colors
['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

# ===============================================================================================================================
# overlapping subplots tick labels
# https://stackoverflow.com/questions/15773049/remove-overlapping-tick-marks-on-subplot-in-matplotlib
from matplotlib.ticker import MaxNLocator # added
nbins = len(ax1.get_xticklabels()) # added 
ax2.yaxis.set_major_locator(MaxNLocator(nbins=nbins, prune='upper')) # added 
# or
ax.xaxis.set_major_locator(plt.MaxNLocator(3))


axes[2].yaxis.set_label_coords(-0.12, -0.05) # , transform = axes[2].transAxes
axes[1].yaxis.set_label_coords(-0.05, -0.05, transform = axes[1].transAxes)

axes[7].set_axis_off()

axes[5].xaxis.set_tick_params(labelbottom=True)

# ===================================================================================================================
ax.tick_params(axis='x', rotation=45)

# ===================================================================================================================================================
lgnd = ax.legend(loc = "lower left", framealpha = 0.1, frameon = True, bbox_to_anchor=(1.01, 0.3))
# lgnd = ax.legend(loc = "lower right", framealpha = 0.1, frameon = True)

#change the marker size manually for both lines
lgnd.legendHandles[0]._sizes = [30]

# ======================================================================================================================================================
# Replace columns values by dict
d = {1: 'winter', 2: 'spring', 3: 'summer', 4: 'autumn'}
df["seasons"] = df["seasons"].map(d)

# ======================================================================================================================================================
# Drop rows all equal zero
df.loc[~(df==0).all(axis=1)]

# ======================================================================================================================================================
# First order derivative to time (seconds)
recv.diff() / recv.index.to_series().diff().dt.total_seconds()

# ======================================================================================================================================================
# Long to wide, change multiindex order and  sort by level
dft = dfo.pivot(index='Flux',columns='Tower')
dft.columns.names = ['Info', 'Tower']
dft = dft.T.swaplevel() # or reorder_levels('Info', 'Tower')

idx, iidx = dft.index.sortlevel(0)

# ======================================================================================================================================================
# add legend by scatter size
import numpy as np
import matplotlib.pyplot as plt

N = 50
x = np.random.rand(N)
y = np.random.rand(N)
a2 = 400*np.random.rand(N)

sc = plt.scatter(x, y, s=a2, alpha=0.5)
plt.legend(*sc.legend_elements("sizes", num=6))
plt.show()


# ======================================================================================================================================================
# Group by every 3 minutes
gps = df.groupby(pd.Grouper(freq = '3T'))

for g in gps.groups:
    print(g)
    gp = gps.get_group(g)
    print(gp)

# ======================================================================================================================================================
# Set matplotlib axis datetime format
import matplotlib.dates as mdates
myFmt = mdates.DateFormatter('%d %H:%M')
axes[0].xaxis.set_major_formatter(myFmt)

# ======================================================================================================================================================
# count values
geolocations['Countrycode'].value_counts()



# ======================================================================================================================================================
#  blended transformation
import matplotlib.transforms as transforms
trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)

# ======================================================================================================================================================
# convert axes / data transform  
# https://stackoverflow.com/questions/62004022/convert-from-data-coordinates-to-axes-coordinates-in-matplotlib

import matplotlib.pyplot as plt

fig, ax = plt.subplots()
# this is in data coordinates
point = (1000, 1000)
trans = ax.transData.transform(point)
trans = ax.transAxes.inverted().transform(trans)
print(ax.get_xlim(), trans)  

ax.set_xlim(0, 2000)
ax.set_ylim(0, 2000)
trans = ax.transData.transform(point)
trans = ax.transAxes.inverted().transform(trans)
print(ax.get_xlim(), trans)

# =======================================================================================================================
# auto nrows and ncols
acnt = len(df.columns) - 1
nc = int(np.ceil(np.sqrt(acnt)))
if nc*(nc-1) >= acnt:
    nr = nc - 1
else:
    nr = nc

fig, axes = setup_canvas(nr,nc, figsize = (5 * nc, 3 * nr))#, wspace = 0.2, hspace = 0.2)

# =======================================================================================================================
# interp nc to sites
df = nc['o3'].interp(longitude = meta['longitude'].to_xarray(), latitude = meta['latitude'].to_xarray()).to_dataframe()

# =======================================================================================================================
# Smooth 2D array
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage import uniform_filter, convolve1d

klen = 2
kernel = np.ones(klen)

arr2Df = arr2D
arr2Df = convolve1d(arr2Df, kernel)
arr2Df = convolve1d(arr2Df.T, kernel).T

nc = nc.coarsen(longitude=2, boundary = 'pad').mean().coarsen(latitude=2, boundary = 'pad').mean()


# mount the drive
def mount_drive(force_remount = False):
    from google.colab import drive
    from pathlib import Path
  
    drive.mount('/content/drive', force_remount = force_remount)
    return Path.cwd().joinpath('drive/My Drive')

def unmount_drive():
    from google.colab import drive
    drive.flush_and_unmount()

# def init_gee(project_name):
#     import ee

#     # Trigger the authentication flow.
#     ee.Authenticate()

#     # Initialize the library.
#     ee.Initialize(project = project_name)

# download from colab
def download_file(src, filename, **kwargs):
    """
    src: data source, dataframe, figures, etc.
    filename: directory to save file (e.g. fig.png)
    """
    import pickle
    from google.colab import files
    # save figures:
    if filename.split(".")[-1] in ["jpeg", "jpg", "png", "pdf"]:
        fig = src
        fig.savefig(filename, bbox_inches = "tight", **kwargs)
    # csv
    elif filename.split(".")[-1] == "csv":
        df = src
        df.to_csv(filename)
    # xlsx
    elif filename.split(".")[-1] == "xlsx":
        df = src
        df.to_excel(filename)
    # pickle
    else:
        with open(filename, "wb") as f:
            pickle.dump(src, f)
    files.download(filename)

def check_colab_gpu():
    # This function is deprecated on 25/10/2022
    import os
    if int(os.environ["COLAB_GPU"]) > 0: # TF_FORCE_GPU_ALLOW_GROWTH: True
        print("A GPU is connected.")
        gpu = 1
    elif "COLAB_TPU_ADDR" in os.environ and os.environ["COLAB_TPU_ADDR"]: # XRT_TPU_CONFIG or COLAB_TPU_ADDR or TPU_NAME in os.environ
        print("A TPU is connected.")
        gpu = 2
    else:
        print("No accelerator is connected.")
        gpu = 0
    return gpu


import numpy as np

# Priestley_Taylor can be used for forests and grasslands for approximation, but with different scaling factor alpha
def Priestley_Taylor(Ta, PA, R, G = 0, alpha = 1.26):
    '''
    Potential evapotranspiration [mm d-1] priestley_assessment_1972
    ETo = alpha * (R-G) * delta_ / (lambda_ * (delta_ + gamma_))
    https://wetlandscapes.github.io/blog/blog/penman-monteith-and-priestley-taylor/

    where is the dryness coefficient, R is the net incoming radiation, G is the heat flux into the ground (R - G =
    LE + H) where H is sensible heat and LE is latent heat, delta_ is the slope of the saturation vapour pressure
    curve, and gamma_ is the psychrometric constant.

    Ta: average day temperature [degC].
    PA: atmospheric pressure [kPa].
    R: net radiation [MJ m-2 d-1]. W m-2 * (1000000/86400) = MJ m-2 d-1
    G: soil heat flux [MJ m-2 d-1].
    '''
    # Latent Heat of Vaporization [MJ kg-1].
    lambda_ = 2.501 - 0.002361 * Ta
    # Specific heat of air [MJ kg-1 degC-1]
    CP = 1.013 * 10**-3
    # Psychrometric constant [kPa degC-1].
    gamma_ = CP * PA / (0.622 * lambda_)

    # Saturation vapor pressure at the air temperature T [kPa]
    es = 0.6108 * np.exp(17.27 * Ta / (Ta + 237.3))
    # Slope of saturation vapour pressure curve at air Temperature [kPa °C-1].
    delta_ = 4098 * es / (Ta + 237.3) ** 2

    ETo = (alpha * delta_ * (R - G)) / (lambda_ * (delta_ + gamma_))
    return ETo


def Priestley_Taylor_JPL(net_radiation, Ta, NDVI):
    """
    Calculate evapotranspiration (ET) using the PT-JPL model.

    Parameters:
    - net_radiation: Net radiation (W / m2)
    - albedo: Surface albedo
    - Ta: Air temperature (degC)
    - NDVI: Normalised Difference Vegetation Index

    Returns:
    - et: Evapotranspiration (mm/day)
    Citation:
    - Fisher, Joshua B., Kevin P. Tu, and Dennis D. Baldocchi. "Global estimates of the land-atmosphere water flux based on monthly AVHRR and ISLSCP-II data, validated at 16 FLUXNET sites." Remote Sensing of Environment 112.3 (2008): 901-919.
    - Wen, Jiaming, et al. "Resolve the clear-sky continuous diurnal cycle of high-resolution ECOSTRESS evapotranspiration and land surface temperature." Water Resources Research 58.9 (2022): e2022WR032227.
    """
    Ta = Ta + 273.15
    # Constants
    lambda_v = 2.45e6  # Latent heat of vaporisation (J/kg)
    gamma = 0.066  # Psychrometric constant (kPa/degC)


    # Vapour pressure deficit (VPD, kPa)
    es = 0.6108 * np.exp((17.27 * (Ta - 273.15)) / ((Ta - 273.15) + 237.3))  # Saturation vapour pressure (kPa)

    # Slope of saturation vapour pressure curve (Delta, kPa/degC)
    delta = 4098 * es / ((Ta - 273.15) + 237.3) ** 2

    # Vegetation constraints (NDVI-based f_veg)
    f_veg = np.clip((NDVI - 0.2) / 0.8, 0, 1)  # Scaling NDVI to [0, 1]


    # Alpha (Priestley-Taylor coefficient)
    alpha = 1.26 * f_veg

    # ET calculation (mm/day)
    et = alpha * delta * net_radiation / (delta + gamma) / lambda_v  # kg/m2/s (convert to mm/day)
    et = et * 86400  # Convert from kg/m2/s to mm/day

    return np.maximum(et, 0)  # Ensure ET is non-negative


def mmday2Wm2(et_mm_day):
    """
    Conversion of mm/day to W/m2:
    1 mm/day = 1 kg/m2/day (since 1 mm of water depth equals 1 kg/m2)
    Latent heat of vaporisation (L_v) = ca. 2.45 MJ/kg = 2.45 x 10^6 J/kg
    Seconds in a day = 24 x 60 x 60 = 86400 s
    Energy flux (W/m2) = (L_v x evaporation rate) / seconds per day
    For 1 mm/day: (2.45 x 10^6 J/kg x 1 kg/m2/day) / 86400 s = ca. 28.4 W/m2

    1 mm/day = 28.4 W/m2
    """
    et_Wm2 = et_mm_day * 28.4
    return et_Wm2




'''
Python code is translated from R version: https://cran.r-project.org/web/packages/bigleaf/bigleaf.pdf
'''

def intercellular_CO2(Ca, GPP, Gs, Rleaf = 0):
    '''
    Ca: Atmospheric or surface CO2 concentration (umol mol-1)
    GPP: Gross primary productivity (umol CO2 m-2 s-1)
    Gs: Surface conductance to water vapor (mol m-2 s-1)
    Rleaf: Ecosystem respiration stemming from leaves (umol CO2 m-2 s-1); defaults to 0
    '''
    Ci = Ca - (GPP - Rleaf) / (Gs / 1.6)
    return Ci

def Rg_to_PPFD(Rg, J_to_mol = 4.6, frac_PAR = 0.5):
    '''
    Rg: Global radiation = incoming short-wave radiation at the surface (W m-2)
    J_to_mol: Conversion factor from J m-2 s-1 (= W m-2) to umol (quanta) m-2 s-1
    frac_PAR: Fraction of incoming solar irradiance that is photosynthetically active radiation (PAR); defaults to 0.5
    PPFD: Photosynthetic photon flux density (umol m-2 s-1)
    '''
    PPFD = Rg * frac_PAR * J_to_mol
    return PPFD

def light_response(NEE, PPFD, Reco):
    from scipy.optimize import curve_fit
    '''
    mod <- nls(-NEE ~ alpha * PPFD / (1 - (PPFD / PPFD_ref) + (alpha * PPFD / GPP_ref)) - Reco,
                start=list(alpha=0.05,GPP_ref=30),...)
    NEE: umol CO2 m-2 s-1
    Reco: umol CO2 m-2 s-1
    PPFD: umol m-2 s-1
    PPFD_ref = 2000 umol m-2 s-1; Falge et al. 2001
    '''
    X = [PPFD, Reco]
    def func(X, alpha, GPP_ref, PPFD_ref = 2000):
        PPFD = X[0]
        Reco = X[1]
        return alpha * PPFD / (1 - (PPFD / PPFD_ref) + (alpha * PPFD / GPP_ref)) - Reco
    popt, pcov = curve_fit(func,  X,  -NEE) 
    return popt, pcov

def latent_heat_vaporization(Tair):
    '''
    Tair Air temperature (deg C)
    λ is Latent heat of vaporization (J kg-1)
    '''
    lambda_ = (2.501 - 0.00237 * Tair) * 10e6
    return lambda_

def LE_to_ET(LE, Tair):
    '''
    LE: Latent heat flux (W m-2)
    Tair: Air temperature (deg C)
    ET: Evapotranspiration (kg m-2 s-1)
    where λ is the latent heat of vaporization (J kg-1) as calculated by latent_heat_vaporization
    '''
    lambda_ = latent_heat_vaporization(Tair)
    ET = LE / lambda_
    return ET

def ET_to_LE(ET, Tair):
    '''
    LE: Latent heat flux (W m-2)
    Tair: Air temperature (deg C)
    ET: Evapotranspiration (kg m-2 s-1)
    where λ is the latent heat of vaporization (J kg-1) as calculated by latent_heat_vaporization
    '''
    lambda_ = latent_heat_vaporization(Tair)
    LE = ET * lambda_
    return LE


#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: Songyan Zhu
# Mail: zhusy93@gmail.com
# Created Time:  2018-10-23 13:28:34
#############################################


from setuptools import setup, find_packages

setup(
	name = "scitbx",
	version = "0.0.80",
	keywords = ("Geosciences, geospatial, data science"),
	description = "For academic data processing and plotting etc.",
	long_description = "Processing academic data and drawing figures for geosciences researchers.",
	license = "MIT Licence",

	url="https://github.com/soonyenju/scitbx",
	author = "Songyan Zhu",
	author_email = "zhusy93@gmail.com",

	packages = find_packages(),
	include_package_data = True,
	platforms = "any",
	install_requires=[

	]
)